"""
This is driven by an FSA that manages the binding and the retrievals.
Initially, (startState) it waits for a binding event.  It then
tryToBind, which uses a timer since it takes time to bind.  If it
succeeds it goes to bindOnState
bindOn turns on the bindingTimer
the bindingTimer turns on bindDone state
bindDone -> start via bindDoneTimer.

Fail states occurs with bindFailState twoPlacesFact, twoObjectsFact,
and notEnoughPlaceObjectsFact.

retrievePlaceState is turned on externally by a query.
It turns off start.  It allows the binding nets to fire, and when a place
fires, it turns on the placeRetrievedFact (extra neurons in the automaton net).
Together these turn on retrievePlaceDone, which then goes back to startState

A similar mechanism is used for retrieveObject.

If you want to unbind, you need to call useUnbinding.
"""
from nealCoverClass import NealCoverFunctions
from timerClass import TimerClass

class cogmapBaseClass:
    numberAutomatonStates = 19 #lastState (below) + 7
    numberObjects = -1
    numberPlaces = -1

    objectToPlaceSynapses = -1
    placeToObjectSynapses = -1
    
    def __init__(self,simName,sim,neal,spinnVersion,fsa):
        self.simName = simName
        self.sim = sim
        self.neal = neal
        self.spinnVersion = spinnVersion
        self.fsa = fsa
        self.automatonCells = self.createAutomaton()
        self.useUnbind = False


    #-----FSA functions and constants
    #state names
    startState = 0
    tryBindState = 1
    bindOnState = 2
    bindFailState = 3
    bindDoneState = 4
    retrieveObjectState = 5
    retrievePlaceState = 6
    retrieveObjectDoneState = 7
    retrievePlaceDoneState = 8
    startUnbindState = 9
    unbindingState = 10
    lastState = 10
    onePlaceOneObjectFact = lastState+1 #one placeBindOn and one objectBindOn on
    twoPlacesFact = lastState+2 #two or more placeBindOn
    twoObjectsFact = lastState+3 #two or more objectBindOn
    notEnoughPlaceObjectsFact = lastState+4 #0 or 1 objects place in total
    placeRetrievedFact = lastState+5
    objectRetrievedFact = lastState+6
    placeNotFoundFact = lastState+7
    objectNotFoundFact = lastState+8

    #Each object and place half connects to the onePlaceOneObjectFact
    #Note that 2 of either will turn it on as will more than one of each.
    #These are handled elsewhere to convert to the bindFailState
    def setupObjectPlaceFacts(self):
        print (self.numberObjects)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                            self.automatonCells,self.onePlaceOneObjectFact)
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                            self.automatonCells,self.twoObjectsFact)

        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                            self.automatonCells,self.onePlaceOneObjectFact)
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                            self.automatonCells,self.twoPlacesFact)

    #---Make the fsa that governs the map.  
    def stateStopsPlaceDone(self,stateNumber):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.placeBindDoneCells,placeNumber)

    def stateStopsObjectDone(self,stateNumber):
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.objectBindDoneCells,objectNumber)

    def stateStopsDone(self,stateNumber):
        self. stateStopsPlaceDone(stateNumber)
        self. stateStopsObjectDone(stateNumber)

    def connectStartState(self):
        #the bind signal moves from start to tryBind
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.startState)
        self.tryToBindTimer.stateStartsTimer(self.automatonCells,
                                                 self.tryBindState)

        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                                       self.automatonCells,self.bindDoneState)

        self.stateStopsDone(self.startState)

        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                            self.automatonCells,self.retrieveObjectDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                            self.automatonCells,self.retrievePlaceDoneState)

        self.statePreventsOutput(self.startState)

        #start state turns off place and objectRetreivedFact
        #and the not found facts.
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                    self.automatonCells,self.objectRetrievedFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                    self.automatonCells,self.placeRetrievedFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                    self.automatonCells,self.placeNotFoundFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                    self.automatonCells,self.objectNotFoundFact)


        #allow movement to the unbindState
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                                    self.automatonCells,self.startUnbindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startUnbindState,
                                       self.automatonCells,self.startState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.startState,
                                       self.automatonCells,self.unbindingState)


    def statePreventsPlaceOutput(self,stateNumber):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.answerPlaceCells,placeNumber)

    def statePreventsObjectOutput(self,stateNumber):
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.answerObjectCells,objectNumber)

    def statePreventsOutput(self,stateNumber):
        self. statePreventsPlaceOutput(stateNumber)
        self. statePreventsObjectOutput(stateNumber)

    #-----Connect States----
    def connectBindOnState(self):
        #1p1o fact moves from tryBind to bindOn
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.bindOnState)
        self.setupObjectPlaceFacts()
        #Half_on is 0.0012, here we go slower so that it comes up slower
        #if two places is on
        funnyWt = 0.001
        if ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 8)):
            funnyWt = 0.0011

        self.fsa.stateStimulatesState(self.automatonCells,
                        self.onePlaceOneObjectFact,self.automatonCells,
                                      self.bindOnState,funnyWt)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindOnState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindOnState,
                                self.automatonCells,self.onePlaceOneObjectFact)
        #stop tryBindTimer
        self.tryToBindTimer.stateStopsTimer(self.automatonCells,
                                            self.bindOnState)

        self.statePreventsOutput(self.bindOnState)
        
    def connectBindFailState(self):
        #2p, 2o, or timer goes to bindFail 
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.tryBindState,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.twoPlacesFact,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.twoObjectsFact,
                                       self.automatonCells,self.bindFailState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.notEnoughPlaceObjectsFact,self.automatonCells,
                                       self.bindFailState)
        self.tryToBindTimer.timerStartsState(self.automatonCells, #timeout fail
                                            self.notEnoughPlaceObjectsFact)

        #fail prevents on
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.bindOnState)

        #turn states and timers off
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.tryBindState)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.twoPlacesFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                                       self.automatonCells,self.twoObjectsFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.onePlaceOneObjectFact)
        self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.notEnoughPlaceObjectsFact)
        self.tryToBindTimer.stateStopsTimer(self.automatonCells,
                                            self.bindFailState)

        #fail goes right back to start
        self.fsa.stateTurnsOnState(self.automatonCells,self.bindFailState,
                            self.automatonCells,self.startState)

    def bindFailStopsPlaceObjectOn(self):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.placeBindOnCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,self.bindFailState,
                            self.objectBindOnCells,objectNumber)

    def bindOnPrimesPlaceAndObjectBind(self):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                            self.placeBindCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                            self.objectBindCells,objectNumber)

    def stopPlaceAndObjectBind(self,stateNumber):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.placeBindCells,placeNumber)
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,stateNumber,
                            self.objectBindCells,objectNumber)

    def connectBindDoneState(self):
        #timer moves to binddone
        self.bindingTimer.stateStartsTimer(self.automatonCells, 
                                            self.bindOnState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.bindOnState,
                                       self.automatonCells,self.bindDoneState)
        self.bindingTimer.timerStartsState(self.automatonCells, 
                                            self.bindDoneState)

        self.bindOnPrimesPlaceAndObjectBind()

        self.fsa.stateTurnsOffState(self.automatonCells,self.bindDoneState,
                            self.automatonCells,self.bindOnState)
        self.stopPlaceAndObjectBind(self.bindDoneState)

        #binding done returns to start via timer
        self.bindDoneTimer.stateStartsTimer(self.automatonCells, 
                                            self.bindDoneState)
        self.bindDoneTimer.timerStartsState(self.automatonCells, 
                                            self.startState)

        self.statePreventsOutput(self.bindDoneState)


    def retrieveObjectPrimesPlaceBind(self):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,
                    self.retrieveObjectState,self.placeBindCells,placeNumber)

    def answerPlaceStartsPlaceRetrievedFact(self):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOnState(self.answerPlaceCells,placeNumber,
                    self.automatonCells,self.placeRetrievedFact)

    def connectRetrieveObjectState(self):
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                        self.automatonCells,self.retrieveObjectState)
        self.retrieveObjectPrimesPlaceBind()
        self.statePreventsPlaceOutput(self.retrieveObjectState)
        self.answerPlaceStartsPlaceRetrievedFact()
        self.fsa.stateTurnsOffState(self.automatonCells,
                self.retrieveObjectState,self.automatonCells,self.startState)

    def connectRetrieveObjectDoneState(self):
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.retrieveObjectState,self.automatonCells,
                                       self.retrieveObjectDoneState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.objectRetrievedFact,self.automatonCells,
                                       self.retrieveObjectDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,
            self.retrieveObjectDoneState,self.automatonCells,
                                       self.retrieveObjectState)

        #Turn off the queryobject rooms.
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.automatonCells,
                        self.retrieveObjectDoneState,
                        self.queryOnPlaceCells,placeNumber)
        
        #don't need the output on anymore
        self.statePreventsObjectOutput(self.retrieveObjectDoneState)
        self.stopPlaceAndObjectBind(self.retrieveObjectDoneState)

        self.fsa.stateTurnsOnState(self.automatonCells,
            self.retrieveObjectDoneState,self.automatonCells,
                                       self.startState)

    def retrievePlacePrimesObjectBind(self):
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.automatonCells,
                    self.retrievePlaceState,self.objectBindCells,objectNumber)

    def answerObjectStartsObjectRetrievedFact(self):
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOnState(self.answerObjectCells,objectNumber,
                    self.automatonCells,self.objectRetrievedFact)

    def connectRetrievePlaceState(self):
        self.fsa.stateHalfTurnsOnState(self.automatonCells,self.startState,
                        self.automatonCells,self.retrievePlaceState)
        self.retrievePlacePrimesObjectBind()
        self.statePreventsObjectOutput(self.retrievePlaceState)
        self.answerObjectStartsObjectRetrievedFact()
        self.fsa.stateTurnsOffState(self.automatonCells,
                self.retrievePlaceState,self.automatonCells,self.startState)

    def connectRetrievePlaceDoneState(self):
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.retrievePlaceState,self.automatonCells,
                                       self.retrievePlaceDoneState)
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
            self.placeRetrievedFact,self.automatonCells,
                                       self.retrievePlaceDoneState)
        self.fsa.stateTurnsOffState(self.automatonCells,
            self.retrievePlaceDoneState,self.automatonCells,
                                       self.retrievePlaceState)

        #Turn off the queryplace objects.
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.automatonCells,
                        self.retrievePlaceDoneState,
                        self.queryOnObjectCells,objectNumber)
            
        #don't need the output on anymore
        self.statePreventsPlaceOutput(self.retrievePlaceDoneState)
        self.stopPlaceAndObjectBind(self.retrievePlaceDoneState)

        self.fsa.stateTurnsOnState(self.automatonCells,
            self.retrievePlaceDoneState,self.automatonCells,
                                       self.startState)

    def unbindingStatePrimesAllBindCells(self):
        anUnbindingNeuronNumber = self.unbindingState*10
        if (self.simName == 'spinnaker') :
            primeWt = 0.002
        elif (self.simName == 'nest') :
            primeWt = 0.003
        for object in range (0,self.numberObjects):
            for toNeuron in range (0,8):
                objectBindCellNumber = (object *10)+toNeuron
                self.fsa.oneNeuronStimulatesOneNeuron(
                    self.automatonCells,anUnbindingNeuronNumber,
                    self.objectBindCells,objectBindCellNumber,primeWt)

        for place in range (0,self.numberPlaces):
            for toNeuron in range (0,8):
                placeBindCellNumber = (place *10)+toNeuron
                self.fsa.oneNeuronStimulatesOneNeuron(
                    self.automatonCells,anUnbindingNeuronNumber,
                    self.placeBindCells,placeBindCellNumber,primeWt)
                

    #Get two neurons, one objectBind, one placeBind to fire.
    #It's not clear which CA is being fired, so send activation to all
    #of the CAs, but the one that is stimulated from the BindOn spikes.
    def unbindTimerNeuronTurnsOnPair(self,fObjectFirst,timer,timerNeuron,
                firstPopulation, firstNeuron, secondPopulation,secondNeuron):
        firstWt = 0.031
        secondWt = 0.035
        #print(firstNeuron,secondNeuron)

        if (fObjectFirst):
            numberFirstCAs = self.numberObjects
            numberSecondCAs = self.numberPlaces
        else:
            numberFirstCAs = self.numberPlaces
            numberSecondCAs = self.numberObjects
            
        for firstCA in range (0,numberFirstCAs):
            neuronToStimulate = firstCA*self.fsa.CA_SIZE+firstNeuron
            self.fsa.oneNeuronStimulatesOneNeuron(timer.timerCells,timerNeuron,
                            firstPopulation,neuronToStimulate,firstWt)
            
        for secondCA in range (0,numberSecondCAs):
            neuronToStimulate = secondCA*self.fsa.CA_SIZE+secondNeuron
            self.fsa.oneNeuronStimulatesOneNeuron(timer.timerCells,timerNeuron,
                            secondPopulation,neuronToStimulate,secondWt)


    def setupUnbinding(self):
        self.unbindingStatePrimesAllBindCells()
        timerNeuron=10

        placeCellOffset = 0
        objectCellOffset = 0

        #Setup spikes so that the object cells fire before the place cells.
        #This makes the o->p weights near 0.
        for placeCellOffset in range (0,8):
            for objectCellOffset in range (0,8):
                self.unbindTimerNeuronTurnsOnPair(False, self.unbindTimer,
                            timerNeuron, self.placeBindCells,placeCellOffset,
                            self.objectBindCells,objectCellOffset)

                timerNeuron=timerNeuron+20
        #Setup spikes so that the place cells fire before the object cells.
        #This makes the p->o weights near 0.
        for objectCellOffset in range (0,8):
            for placeCellOffset in range (0,8):
                self.unbindTimerNeuronTurnsOnPair(True, self.unbindTimer,
                            timerNeuron, self.objectBindCells,objectCellOffset,
                            self.placeBindCells,placeCellOffset)


                timerNeuron=timerNeuron+20
        """
        """
        print ("unbind")
        
    def connectUnbindState(self):
        #undone
        #self.unbindTimer.stateStartsTimer(self.automatonCells, 
        #                                    self.startUnbindState)
        unbindStateNeuronNumber = self.startUnbindState*10
        self.fsa.oneNeuronStimulatesOneNeuron(self.automatonCells,
                unbindStateNeuronNumber, self.unbindTimer.timerCells,0,0.03)

        #switch to the unbinding state when the timer starts
        self.fsa.stateHalfTurnsOnState(self.automatonCells,
                self.startUnbindState,self.automatonCells,self.unbindingState)

        self.fsa.oneNeuronTurnsOnState(self.unbindTimer.timerCells,0,
                                    self.automatonCells,self.unbindingState)
        
        self.fsa.stateTurnsOffState(self.automatonCells,self.unbindingState,
                                    self.automatonCells,self.startUnbindState)

        #when the timer is done, switch back to start.
        lastTimerCell = (self.unbindTimer.numberStates*self.fsa.CA_SIZE)-1
        self.fsa.oneNeuronTurnsOnState(self.unbindTimer.timerCells,
                lastTimerCell,self.automatonCells,self.startState)

        self.setupUnbinding()

    #To allow unbinding of objects and places, call this.
    def useUnbinding(self):
        self.useUnbind = True

        #allocate the timer
        self.unbindTimer = TimerClass(self.simName,self.sim,self.neal,
                                self.spinnVersion,self.fsa,257)
        self.unbindTimer.makeFastTimerSynapses()

        self.connectUnbindState()

    #When a query starts, a timer starts 
    #If the query is not done by the end of the timer
    #it fails (turning on the appropriate fact).
    def connectRetrieveObjectFailed(self):
        self.queryObjectTimer.stateStartsTimer(self.automatonCells, 
                                            self.retrieveObjectState)
        self.queryObjectTimer.timerStartsState(self.automatonCells, 
                                            self.retrieveObjectDoneState)
        self.queryObjectTimer.timerStartsState(self.automatonCells, 
                                            self.objectNotFoundFact)

        #both start and retrieveObjectDone stops the timer so the
        #object is found.
        self.queryObjectTimer.stateStopsTimer(self.automatonCells,
                                            self.retrieveObjectDoneState)
        
    #When a query starts, a timer starts 
    #If the query is not done by the end of the timer
    #it fails (turning on the appropriate fact).
    def connectRetrievePlaceFailed(self):
        self.queryPlaceTimer.stateStartsTimer(self.automatonCells, 
                                            self.retrievePlaceState)
        self.queryPlaceTimer.timerStartsState(self.automatonCells, 
                                            self.retrievePlaceDoneState)
        self.queryPlaceTimer.timerStartsState(self.automatonCells, 
                                            self.placeNotFoundFact)

        #both start and retrievePlaceDone stops the timer so the
        #object is found.
        self.queryPlaceTimer.stateStopsTimer(self.automatonCells,
                                            self.retrievePlaceDoneState)
        
    def connectAutomaton(self):
        self.connectStartState()
        self.connectBindOnState()
        self.connectBindFailState()
        self.bindFailStopsPlaceObjectOn()
        self.connectBindDoneState()
        self.connectRetrieveObjectState()
        self.connectRetrievePlaceState()
        self.connectRetrieveObjectDoneState()
        self.connectRetrievePlaceDoneState()
        self.connectRetrieveObjectFailed()
        self.connectRetrievePlaceFailed()

    def createTimers(self):
        self.tryToBindTimer = TimerClass(self.simName,self.sim,self.neal,
                                         self.spinnVersion,self.fsa,10)
        self.bindingTimer = TimerClass(self.simName,self.sim,self.neal,
                                       self.spinnVersion,self.fsa,10)
        self.bindDoneTimer = TimerClass(self.simName,self.sim,self.neal,
                                        self.spinnVersion,self.fsa,6)
        self.queryObjectTimer = TimerClass(self.simName,self.sim,self.neal,
                                           self.spinnVersion,self.fsa,10)
        self.queryPlaceTimer = TimerClass(self.simName,self.sim,self.neal,
                                           self.spinnVersion,self.fsa,10)

        
        #the timerClass sets recording
        self.tryToBindTimer.makeStopNoRestartTimerSynapses()
        self.bindingTimer.makeStopNoRestartTimerSynapses()
        self.bindDoneTimer.makeStopNoRestartTimerSynapses()
        self.queryObjectTimer.makeStopNoRestartTimerSynapses()
        self.queryPlaceTimer.makeStopNoRestartTimerSynapses()

    def createAutomaton(self):
        numNeurons = self.fsa.CA_SIZE * self.numberAutomatonStates
        
        cells=self.sim.Population(numNeurons,self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        for stateNumber in range (0,self.numberAutomatonStates):
            self.fsa.makeCA(cells,stateNumber)

        self.createTimers()
        return cells

    def createObjects(self, numberObjects):
        self.numberObjects = numberObjects
        #print ('objects',numberObjects)
        numberObjectCells = self.numberObjects*self.fsa.CA_SIZE
        self.objectBindCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.objectBindOnCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.objectBindDoneCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.queryOnObjectCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.answerObjectCells = self.sim.Population(numberObjectCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)

        for objectNumber in range (0,self.numberObjects):
            self.fsa.makeCA(self.objectBindCells,objectNumber)
            self.fsa.makeCA(self.objectBindOnCells,objectNumber)
            self.fsa.makeCA(self.objectBindDoneCells,objectNumber)
            self.fsa.makeCA(self.queryOnObjectCells,objectNumber)
            self.fsa.makeCA(self.answerObjectCells,objectNumber)

    def connectObjects(self):
        for objectNumber in range (0,self.numberObjects):
            self.fsa.stateHalfTurnsOnState(self.objectBindOnCells,objectNumber,
                                       self.objectBindCells,objectNumber)
            self.fsa.stateTurnsOnState(self.objectBindCells,objectNumber,
                                       self.objectBindDoneCells,objectNumber)
            self.fsa.stateTurnsOffState(self.objectBindDoneCells,objectNumber,
                                       self.objectBindOnCells,objectNumber)
            self.fsa.stateHalfTurnsOnState(self.queryOnObjectCells,objectNumber,
                                       self.objectBindCells,objectNumber)
            self.fsa.stateTurnsOnState(self.objectBindCells,objectNumber,
                                       self.answerObjectCells,objectNumber)

    #Create neurons for each place and make them CAs.
    def createPlaces(self, numberPlaces):
        self.numberPlaces = numberPlaces
        print ('places',numberPlaces)
        numberPlaceCells = self.numberPlaces*self.fsa.CA_SIZE
        self.placeBindCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.placeBindOnCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.placeBindDoneCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.queryOnPlaceCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)
        self.answerPlaceCells = self.sim.Population(numberPlaceCells,
                self.sim.IF_cond_exp,self.fsa.CELL_PARAMS)

        for placeNumber in range (0,self.numberPlaces):
            self.fsa.makeCA(self.placeBindCells,placeNumber)
            self.fsa.makeCA(self.placeBindOnCells,placeNumber)
            self.fsa.makeCA(self.placeBindDoneCells,placeNumber)
            self.fsa.makeCA(self.queryOnPlaceCells,placeNumber)
            self.fsa.makeCA(self.answerPlaceCells,placeNumber)

    def connectPlaces(self):
        for placeNumber in range (0,self.numberPlaces):
            self.fsa.stateHalfTurnsOnState(self.placeBindOnCells,placeNumber,
                                       self.placeBindCells,placeNumber)
            self.fsa.stateTurnsOnState(self.placeBindCells,placeNumber,
                                       self.placeBindDoneCells,placeNumber)
            self.fsa.stateTurnsOffState(self.placeBindDoneCells,placeNumber,
                                       self.placeBindOnCells,placeNumber)
            self.fsa.stateHalfTurnsOnState(self.queryOnPlaceCells,placeNumber,
                                       self.placeBindCells,placeNumber)
            self.fsa.stateTurnsOnState(self.placeBindCells,placeNumber,
                                       self.answerPlaceCells,placeNumber)

    def setupCogMapRecording(self):
        if ((self.simName == 'spinnaker') and (self.spinnVersion == 7)):
            self.placeBindCells.record()
            self.objectBindCells.record()
            self.automatonCells.record()
            self.placeBindOnCells.record()
            self.objectBindOnCells.record()
            self.placeBindDoneCells.record()
            self.objectBindDoneCells.record()
            self.queryOnPlaceCells.record()
            self.queryOnObjectCells.record()
            self.answerPlaceCells.record()
            self.answerObjectCells.record()

        elif  ((self.neal.simulator == 'nest') or 
             ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 8))):
            self.placeBindCells.record({'spikes'})
            self.objectBindCells.record({'spikes'})
            self.automatonCells.record({'spikes'})
            self.placeBindOnCells.record({'spikes'})
            self.objectBindOnCells.record({'spikes'})
            self.placeBindDoneCells.record({'spikes'})
            self.objectBindDoneCells.record({'spikes'})
            self.queryOnPlaceCells.record({'spikes'})
            self.queryOnObjectCells.record({'spikes'})
            self.answerPlaceCells.record({'spikes'})
            self.answerObjectCells.record({'spikes'})

    def getWellConnectedConn(self,fromSize,toSize,weight):
        connector = []
        for fromNeuron in range (0,fromSize):
            for toNeuron in range (0, toSize):
                connector=connector+[(fromNeuron,toNeuron,weight,
                                      self.neal.DELAY)]
        return connector

    #Make a list connector so all of the excitatory neurons from the CAs in
    #one population connect to the CAs in another population (with weight 0).
    def getBindCAsConnector(self,numberFromCAs,numberToCAs):
        connector = []
        weight = 0.0
        for fromCA in range (0,numberFromCAs):
            for toCA in range (0,numberToCAs):
                for fromOffset in range (0,8):
                    fromNeuronNum = (fromCA*10) + fromOffset
                    for toOffset in range (0,8):
                        toNeuronNum = (toCA*10) + toOffset
                        connector=connector+[(
                            fromNeuronNum,toNeuronNum,weight, self.neal.DELAY)]

        #print (connector)
        bindCAsConnList = self.sim.FromListConnector(connector)
        return bindCAsConnList

    def makeLearningSynapses(self):
        if (self.simName == "nest"):
            #tauP = 16.7; tauM = 3.7;AP = 0.005; AM = 0.002; W_MAX=0.02
            tauP = 10.0; tauM = 10.0;AP = 0.05; AM = 0.1; W_MAX=0.02
            stdp_model = self.sim.STDPMechanism(
                timing_dependence=self.sim.SpikePairRule(
                    tau_plus=tauP, tau_minus=tauM,A_plus=AP,A_minus=AM),
                    weight_dependence=self.sim.AdditiveWeightDependence(
                        w_min=0, w_max=W_MAX),
                    weight=0.0,delay=self.neal.DELAY,
                dendritic_delay_fraction=float(1))

            objectToPlaceListConnector = self.getBindCAsConnector(
                self.numberObjects,self.numberPlaces)
            self.objectToPlaceSynapses = self.sim.Projection(
                    self.objectBindCells,self.placeBindCells,
                               objectToPlaceListConnector,
                               synapse_type = stdp_model)

            placeToObjectListConnector = self.getBindCAsConnector(
                self.numberPlaces,self.numberObjects)
            self.placeToObjectSynapses = self.sim.Projection(
                    self.placeBindCells,self.objectBindCells,
                               placeToObjectListConnector,
                               synapse_type = stdp_model)

        elif ((self.simName == 'spinnaker') and (self.spinnVersion == 8)):
            #tauP = 10.0; tauM = 10.0;AP = 0.05; AM = 0.1; W_MAX=0.02
            tauP = 10.0; tauM = 10.0;AP = 0.05; AM = 0.1; W_MAX=0.02
            stdp_model = self.sim.STDPMechanism(
                    #timing_dependence = self.sim.SpikePairRule(tau_plus=16.0, 
                    #                tau_minus=3.0,A_plus=0.002, A_minus=0.0),
                    #weight_dependence = self.sim.AdditiveWeightDependence(
                    #    w_min=0.0, w_max=0.02) 
                    timing_dependence = self.sim.SpikePairRule(tau_plus=tauP, 
                                    tau_minus=tauM,A_plus=AP, A_minus=AM),
                    weight_dependence = self.sim.AdditiveWeightDependence(
                        w_min=0.0, w_max=W_MAX) 
                )

            placeToObjectListConnector = self.getBindCAsConnector(
                self.numberPlaces,self.numberObjects)
            self.placeToObjectSynapses = self.sim.Projection(
                self.placeBindCells,self.objectBindCells,
                                placeToObjectListConnector,
                                synapse_type=stdp_model)

            objectToPlaceListConnector = self.getBindCAsConnector(
                self.numberObjects,self.numberPlaces)
            self.objectToPlaceSynapses = self.sim.Projection(
                self.objectBindCells,self.placeBindCells,
                objectToPlaceListConnector, synapse_type=stdp_model)
            #print 'bob', placeToObjectConnector
            #print objectToPlaceConnector


        else:
                print ('bad simulator in cogmap.makeLearningSynapses', nealParameters.simulator)
        

    #---Defs for turning things on and off with spike sources.
    def sourceTurnsOnPlace(self, placeNumber, source):
        self.fsa.turnOnStateFromSpikeSource(source,self.placeBindCells,
                                            placeNumber)
    def sourceTurnsOffPlace(self, placeNumber, source):
        self.fsa.turnOffStateFromSpikeSource(source,self.placeBindCells,
                                            placeNumber)

    def sourceTurnsOnObject(self, objectNumber, source):
        self.fsa.turnOnStateFromSpikeSource(source,self.objectBindCells,
                                            objectNumber)
    def sourceTurnsOffObject(self, objectNumber, source):
        self.fsa.turnOffStateFromSpikeSource(source,self.objectBindCells,
                                            objectNumber)

    def sourceTurnsOnBind(self, source):
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.tryBindState)

    def sourceTurnsOnUnbind(self, source):
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.startUnbindState)

    def sourceTurnsOnObjectOn(self, objectNumber, source):
        self.fsa.turnOnStateFromSpikeSource(source,self.objectBindOnCells,
                                            objectNumber)

    def sourceTurnsOnPlaceOn(self, placeNumber, source):
        self.fsa.turnOnStateFromSpikeSource(source,self.placeBindOnCells,
                                            placeNumber)

    def sourceStartsAutomaton(self, source):
        self.fsa.turnOnStateFromSpikeSource(source,self.automatonCells,
                                             self.startState)

    def sourceTurnOnRetrieveObjectFromPlace(self, source):
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.retrieveObjectState)

    def sourceTurnOnRetrievePlaceFromObject(self, source):
        self.fsa.halfTurnOnStateFromSpikeSource(source,self.automatonCells,
                                                self.retrievePlaceState)

    def sourceTurnsOnPlaceQuery(self, source,placeNumber):
        self.fsa.turnOnStateFromSpikeSource(source,self.queryOnPlaceCells,
                                            placeNumber)

    def sourceTurnsOnObjectQuery(self, source,objectNumber):
        self.fsa.turnOnStateFromSpikeSource(source,self.queryOnObjectCells,
                                            objectNumber)

    #---Defs for turning things on and off with states.
    def stateTurnsOnObjectQuery(self,cells,stateNumber,objectNumber):
        self.fsa.stateHalfTurnsOnState(cells,stateNumber,
                                  self.automatonCells, self.retrievePlaceState)
        self.fsa.stateTurnsOnState(cells,stateNumber,
                                   self.queryOnObjectCells,objectNumber)


    #All of the object queries turn off the state
    def objectQueryAnswerTurnsOffState(self,cells,stateNumber):
        for place in range (0,self.numberPlaces):
            self.fsa.stateTurnsOffState(self.answerPlaceCells,place,
                                    cells,stateNumber)

        #The no place answer also turns it off
        self.fsa.stateTurnsOffState(self.automatonCells,self.placeNotFoundFact,
                                    cells,stateNumber)


        

    def stateTurnsOnPlaceQuery(self,cells,stateNumber,placeNumber):
        self.fsa.stateHalfTurnsOnState(cells,stateNumber,
                                  self.automatonCells, self.retrieveObjectState)
        self.fsa.stateTurnsOnState(cells,stateNumber,
                                   self.queryOnPlaceCells,placeNumber)

    def placeQueryAnswerTurnsOffState(self,cells,stateNumber):
        for answer in range (0,self.numberObjects):
            self.fsa.stateTurnsOffState(self.answerObjectCells,answer,
                                    cells,stateNumber)

        #The no place answer also turns it off
        self.fsa.stateTurnsOffState(self.automatonCells,self.objectNotFoundFact,
                                    cells,stateNumber)

        #undone
            

    #--defs for getting populations and synapses.  Cover functions
    def getObjectAnswerCells(self):
        return self.answerObjectCells

    def getPlaceAnswerCells(self):
        return self.answerPlaceCells

    def getAutomatonCells(self):
        return self.automatonCells

    #for debugging
    def getObjectToPlaceSynapses(self):
        return self.objectToPlaceSynapses

    #for debuggging
    def getPlaceToObjectSynapses(self):
        return self.placeToObjectSynapses

    #for debuggging
    def getPlaceBindCells(self):
        return self.placeBindCells

    #for debuggging
    def getObjectBindCells(self):
        return self.objectBindCells

    #-- Save the spike files.
    def printCogMapNets(self):
        if  ((self.neal.simulator == 'nest') or 
             ((self.neal.simulator == 'spinnaker') and (self.neal.spinnVersion == 8))):
                suffix = '.pkl'
        self.placeBindCells.write_data('results/cmPlaceBind'+suffix)
        self.objectBindCells.write_data('results/cmObjectBind'+suffix)
        self.automatonCells.write_data('results/cmAutomaton'+suffix)
        self.placeBindOnCells.write_data('results/cmPlaceBindOn'+suffix)
        self.objectBindOnCells.write_data('results/cmObjectBindOn'+suffix)
        self.placeBindDoneCells.write_data('results/cmPlaceBindDone'+suffix)
        self.objectBindDoneCells.write_data('results/cmObjectBindDone'+suffix)
        self.queryOnPlaceCells.write_data('results/cmQueryOnPlace'+suffix)
        self.queryOnObjectCells.write_data('results/cmQueryOnObject'+suffix)
        self.answerPlaceCells.write_data('results/cmAnswerPlace'+suffix)
        self.answerObjectCells.write_data('results/cmAnswerObject'+suffix)
        self.tryToBindTimer.printResults('results/cmTryToBindTimer')
        self.bindingTimer.printResults('results/cmBindingTimer')
        self.bindDoneTimer.printResults('results/cmBindDoneTimer')
        self.queryObjectTimer.printResults('results/cmObjectQueryTimer')
        self.queryPlaceTimer.printResults('results/cmPlaceQueryTimer')
        if (self.useUnbind):
            self.unbindTimer.printResults('results/cmUnbindTimer')
